/*
 * Copyright (C) 2008 Martin Willi
 * Hochschule fuer Technik Rapperswil
 * Copyright (C) 2005 David Zeuthen
 * Copyright (C) 2005-2008 Dan Williams
 *
 * Based on NetworkManager's vpnc plugin
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib.h>
#include <glib/gi18n-lib.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#define NM_VPN_API_SUBJECT_TO_CHANGE

#include <libnm-glib/nm-settings.h>

#include <nm-vpn-plugin-ui-interface.h>
#include <nm-setting-vpn.h>
#include <nm-setting-connection.h>
#include <nm-setting-ip4-config.h>

#include "nm-openswan.h"

#include "src/nm-openswan-service.h"
#include "common-gnome/keyring-helpers.h"

#define OPENSWAN_PLUGIN_NAME    _("Openswan IPsec, IPsec/xL2TP")
#define OPENSWAN_PLUGIN_DESC    _("Open Source implementation of IPsec")
#define OPENSWAN_PLUGIN_SERVICE NM_DBUS_SERVICE_OPENSWAN


/************** plugin class **************/

static void openswan_plugin_ui_interface_init (NMVpnPluginUiInterface *iface_class);

G_DEFINE_TYPE_EXTENDED (OpenswanPluginUi, openswan_plugin_ui, G_TYPE_OBJECT, 0,
						G_IMPLEMENT_INTERFACE (NM_TYPE_VPN_PLUGIN_UI_INTERFACE,
											   openswan_plugin_ui_interface_init))

/************** UI widget class **************/

static void openswan_plugin_ui_widget_interface_init (NMVpnPluginUiWidgetInterface *iface_class);

G_DEFINE_TYPE_EXTENDED (OpenswanPluginUiWidget, openswan_plugin_ui_widget, G_TYPE_OBJECT, 0,
						G_IMPLEMENT_INTERFACE (NM_TYPE_VPN_PLUGIN_UI_WIDGET_INTERFACE,
											   openswan_plugin_ui_widget_interface_init))

#define OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), OPENSWAN_TYPE_PLUGIN_UI_WIDGET, OpenswanPluginUiWidgetPrivate))

typedef struct {
	GladeXML *xml;
	GtkWidget *widget;
	GtkWindowGroup *window_group;
	GtkSizeGroup *left_size_group;
	GtkSizeGroup *right_size_group;	
	GHashTable *advanced;
	gboolean window_added;
	
	gboolean supports_nss;
} OpenswanPluginUiWidgetPrivate;


#define OPENSWAN_PLUGIN_UI_ERROR openswan_plugin_ui_error_quark ()

static GQuark
openswan_plugin_ui_error_quark (void)
{
	static GQuark error_quark = 0;

	if (G_UNLIKELY (error_quark == 0))
		error_quark = g_quark_from_static_string ("openswan-plugin-ui-error-quark");

	return error_quark;
}

#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }

GType
openswan_plugin_ui_error_get_type (void)
{
	static GType etype = 0;

	if (etype == 0) {
		static const GEnumValue values[] = {
			/* Unknown error. */
			ENUM_ENTRY (OPENSWAN_PLUGIN_UI_ERROR_UNKNOWN, "UnknownError"),
			/* The specified property was invalid. */
			ENUM_ENTRY (OPENSWAN_PLUGIN_UI_ERROR_INVALID_PROPERTY, "InvalidProperty"),
			/* The specified property was missing and is required. */
			ENUM_ENTRY (OPENSWAN_PLUGIN_UI_ERROR_MISSING_PROPERTY, "MissingProperty"),			
			/* The connection was missing invalid. */
			ENUM_ENTRY (OPENSWAN_PLUGIN_UI_ERROR_INVALID_CONNECTION, "InvalidConnection"),
			{ 0, 0, 0 }
		};
		etype = g_enum_register_static ("OpenswanPluginUiError", values);
	}
	return etype;
}

static GList* get_ipsec_connections_list(gboolean x_nm)
{
    gchar *argv[] = {"/usr/bin/python", NM_OPENSWAN_SERVICE_DIR"/properties_utils.py",
                     x_nm ? "--list-ipsec-connections-xnm" : "--list-ipsec-connections",
                     NULL};
    gchar *envs[] = { NULL };
    gchar *wd = ".";
    GError *error = NULL;
    gint sout;
    GPid child_pid;
    if(g_spawn_async_with_pipes(wd, argv, envs, 0, NULL, NULL,
                                &child_pid, NULL, &sout, NULL,
                                &error)){
        GList* result = NULL;
        char buf;
        GString* strres = g_string_new("");
        while(read(sout, &buf, 1) > 0){
            if(buf != '\n'){
                strres = g_string_append_c(strres, buf);
            } else {
                result = g_list_append(result, strres);
                strres = g_string_new("");
            }
        }
        close(sout);
        g_spawn_close_pid(child_pid);
        return result;
    }
    return NULL;
}

static GList* get_connection_info(gchar *name)
{
    gchar *argv[] = {"/usr/bin/python", NM_OPENSWAN_SERVICE_DIR"/properties_utils.py",
                     "--connection-info", name,
                     NULL};
    gchar *envs[] = { NULL };
    gchar *wd = ".";
    GError *error = NULL;
    gint sout;
    GPid child_pid;
    if(g_spawn_async_with_pipes(wd, argv, envs, 0, NULL, NULL,
                                &child_pid, NULL, &sout, NULL,
                                &error)){
        GList* result = NULL;
        char buf;
        GString* strres = g_string_new("");
        while(read(sout, &buf, 1) > 0){
            if(buf != '\n'){
                strres = g_string_append_c(strres, buf);
            } else {
                result = g_list_append(result, strres);
                strres = g_string_new("");
            }
        }
        close(sout);
        g_spawn_close_pid(child_pid);
        return result;
    }
    return NULL;
}

static gboolean
check_validity (OpenswanPluginUiWidget *self, GError **error)
{
	return TRUE;
}

static void update_layout (GtkWidget *widget, OpenswanPluginUiWidgetPrivate *priv)
{
    g_printf("Update ... ");
    if (widget == glade_xml_get_widget (priv->xml, "type-cb")) {
        g_printf("type-cb\n");
        int active = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
        if (active == 1) { // Host-to-Network
            gtk_widget_set_sensitive(glade_xml_get_widget (priv->xml, "remotenetwork-entry"), TRUE);
            gtk_widget_set_sensitive(glade_xml_get_widget (priv->xml, "mask-entry"), TRUE);
            gtk_widget_set_sensitive(glade_xml_get_widget (priv->xml, "gateway-entry"), TRUE); 
            gtk_widget_show(glade_xml_get_widget (priv->xml, "advanced-button"));
            gtk_widget_show(glade_xml_get_widget (priv->xml, "auth-vbox"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "connname-label"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "connname-cb"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "xnm-chb"));
            gtk_widget_show(glade_xml_get_widget (priv->xml, "gateway-entry"));            
            gtk_widget_show(glade_xml_get_widget (priv->xml, "remotenetwork-entry"));
            gtk_widget_show(glade_xml_get_widget (priv->xml, "gateway-label"));            
            gtk_widget_show(glade_xml_get_widget (priv->xml, "remotenetwork-label"));
            gtk_widget_show(glade_xml_get_widget (priv->xml, "mask-entry"));
            gtk_widget_show(glade_xml_get_widget (priv->xml, "mask-label"));
        } else if (active == 2) { // Manual-Ipsec
            gtk_widget_show(glade_xml_get_widget (priv->xml, "connname-label"));
            gtk_widget_show(glade_xml_get_widget (priv->xml, "connname-cb"));
            gtk_widget_show(glade_xml_get_widget (priv->xml, "xnm-chb"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "auth-vbox"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "advanced-button"));
            //gtk_widget_set_sensitive(glade_xml_get_widget (priv->xml, "gateway-entry"), FALSE);            
            //gtk_widget_set_sensitive(glade_xml_get_widget (priv->xml, "remotenetwork-entry"), FALSE);
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "gateway-entry"));            
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "remotenetwork-entry"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "gateway-label"));            
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "remotenetwork-label"));            
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "mask-entry"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "mask-label"));           
        } else { // Host-to-Host
            active = 0;
            gtk_combo_box_set_active (GTK_COMBO_BOX (widget), active);
            gtk_widget_set_sensitive(glade_xml_get_widget (priv->xml, "remotenetwork-entry"), FALSE);
            gtk_widget_set_sensitive(glade_xml_get_widget (priv->xml, "mask-entry"), FALSE);  
            gtk_widget_set_sensitive(glade_xml_get_widget (priv->xml, "gateway-entry"), TRUE); 
            gtk_widget_show(glade_xml_get_widget (priv->xml, "advanced-button"));
            gtk_widget_show(glade_xml_get_widget (priv->xml, "auth-vbox"));          
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "connname-label"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "connname-cb"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "xnm-chb"));      
            gtk_widget_show(glade_xml_get_widget (priv->xml, "gateway-entry"));            
            gtk_widget_show(glade_xml_get_widget (priv->xml, "remotenetwork-entry"));
            gtk_widget_show(glade_xml_get_widget (priv->xml, "gateway-label"));            
            gtk_widget_show(glade_xml_get_widget (priv->xml, "remotenetwork-label"));  
            gtk_widget_show(glade_xml_get_widget (priv->xml, "mask-entry")); 
            gtk_widget_show(glade_xml_get_widget (priv->xml, "mask-label"));   
        }
    } else if (widget == glade_xml_get_widget (priv->xml, "auth-cb")) {
        int active = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
        g_printf("auth-cb\n");
        if(priv->supports_nss){
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "cert-label"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "cert-filechooser"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "rsa-label"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "rsa-filechooser"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "password-label"));
            gtk_widget_hide(glade_xml_get_widget (priv->xml, "password-entry"));
            return;
        }
        switch(active) {
        case 1: // Raw RSA                
                gtk_widget_hide(glade_xml_get_widget (priv->xml, "cert-label"));
                gtk_widget_hide(glade_xml_get_widget (priv->xml, "cert-filechooser"));
                gtk_widget_show(glade_xml_get_widget (priv->xml, "rsa-label"));
                gtk_widget_show(glade_xml_get_widget (priv->xml, "rsa-filechooser"));
                gtk_widget_hide(glade_xml_get_widget (priv->xml, "password-label"));
                gtk_widget_hide(glade_xml_get_widget (priv->xml, "password-entry"));
                break;
        case 2: // PSK
                gtk_widget_hide(glade_xml_get_widget (priv->xml, "cert-label"));
                gtk_widget_hide(glade_xml_get_widget (priv->xml, "cert-filechooser"));
                gtk_widget_hide(glade_xml_get_widget (priv->xml, "rsa-label"));
                gtk_widget_hide(glade_xml_get_widget (priv->xml, "rsa-filechooser"));
                gtk_widget_show(glade_xml_get_widget (priv->xml, "password-label"));
                gtk_widget_show(glade_xml_get_widget (priv->xml, "password-entry"));                         
                break;
        default:// X.509
                active = 0;
                gtk_combo_box_set_active(GTK_COMBO_BOX (widget), active);
                gtk_widget_show(glade_xml_get_widget (priv->xml, "cert-label"));
                gtk_widget_show(glade_xml_get_widget (priv->xml, "cert-filechooser"));
                gtk_widget_hide(glade_xml_get_widget (priv->xml, "rsa-label"));
                gtk_widget_hide(glade_xml_get_widget (priv->xml, "rsa-filechooser"));
                gtk_widget_show(glade_xml_get_widget (priv->xml, "password-label"));
                gtk_widget_show(glade_xml_get_widget (priv->xml, "password-entry"));
        }
        
    }
}

static void
settings_changed_cb (GtkWidget *widget, gpointer user_data)
{
	OpenswanPluginUiWidget *self = OPENSWAN_PLUGIN_UI_WIDGET (user_data);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	
	g_signal_emit_by_name (OPENSWAN_PLUGIN_UI_WIDGET (user_data), "changed");
}

static void
type_changed_cb (GtkWidget *widget, gpointer user_data)
{
	OpenswanPluginUiWidget *self = OPENSWAN_PLUGIN_UI_WIDGET (user_data);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	
	update_layout(widget, priv);
	g_signal_emit_by_name (OPENSWAN_PLUGIN_UI_WIDGET (user_data), "changed");
}

static void
auth_changed_cb (GtkWidget *widget, gpointer user_data)
{
	OpenswanPluginUiWidget *self = OPENSWAN_PLUGIN_UI_WIDGET (user_data);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	
	update_layout(widget, priv);
	g_signal_emit_by_name (OPENSWAN_PLUGIN_UI_WIDGET (user_data), "changed");
}

static void
stuff_changed_cb (GtkWidget *widget, gpointer user_data)
{
	g_signal_emit_by_name (OPENSWAN_PLUGIN_UI_WIDGET (user_data), "changed");
}

static void
advanced_dialog_close_cb (GtkWidget *dialog, gpointer user_data)
{
	gtk_widget_hide (dialog);
	/* gtk_widget_destroy() will remove the window from the window group */
	gtk_widget_destroy (dialog);
}

static void
advanced_dialog_response_cb (GtkWidget *dialog, gint response, gpointer user_data)
{
	OpenswanPluginUiWidget *self = OPENSWAN_PLUGIN_UI_WIDGET (user_data);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	GError *error = NULL;

	if (response != GTK_RESPONSE_OK) {
		advanced_dialog_close_cb (dialog, self);
		return;
	}

	if (priv->advanced)
		g_hash_table_destroy (priv->advanced);
	priv->advanced = advanced_dialog_new_hash_from_dialog (dialog, &error);
	if (!priv->advanced) {
		g_message ("%s: error reading advanced settings: %s", __func__, error->message);
		g_error_free (error);
	}
	advanced_dialog_close_cb (dialog, self);

	stuff_changed_cb (NULL, self);
}

static void
advanced_button_clicked_cb (GtkWidget *button, gpointer user_data)
{
	OpenswanPluginUiWidget *self = OPENSWAN_PLUGIN_UI_WIDGET (user_data);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	GtkWidget *dialog, *toplevel;

	toplevel = gtk_widget_get_toplevel (priv->widget);
	g_return_if_fail (GTK_WIDGET_TOPLEVEL (toplevel));

	dialog = advanced_dialog_new (priv->advanced);
	if (!dialog) {
		g_warning ("%s: failed to create the Advanced dialog!", __func__);
		return;
	}

	gtk_window_group_add_window (priv->window_group, GTK_WINDOW (dialog));
	if (!priv->window_added) {
		gtk_window_group_add_window (priv->window_group, GTK_WINDOW (toplevel));
		priv->window_added = TRUE;
	}

	gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (toplevel));
	g_signal_connect (G_OBJECT (dialog), "response", G_CALLBACK (advanced_dialog_response_cb), self);
	g_signal_connect (G_OBJECT (dialog), "close", G_CALLBACK (advanced_dialog_close_cb), self);

	gtk_widget_show_all (dialog);
	
}

static gboolean
is_openswan_support_nss ()
{
    return ENABLE_NSS;
}

static void
show_toggled_cb (GtkCheckButton *button, OpenswanPluginUiWidget *self)
{
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	GtkWidget *widget;
	gboolean visible;

	visible = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button));

	widget = glade_xml_get_widget (priv->xml, "password-entry");
	g_assert (widget);
	gtk_entry_set_visibility (GTK_ENTRY (widget), visible);
}

static void
load_connections_list (OpenswanPluginUiWidget *self, const char *value)
{
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	GtkWidget *widget;
	gboolean xnm;

	xnm = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (glade_xml_get_widget (priv->xml, "xnm-chb")));
	// reload
	widget = glade_xml_get_widget (priv->xml, "connname-cb");
	g_assert (widget);
    
    //clear
    gtk_combo_box_set_active(GTK_COMBO_BOX (widget), 0);
    while(gtk_combo_box_get_active(GTK_COMBO_BOX (widget)) >= 0){       
        gtk_combo_box_remove_text(GTK_COMBO_BOX (widget), 0);
        gtk_combo_box_set_active(GTK_COMBO_BOX (widget), 0);
    }
    
    //load
	GList* conn_list = get_ipsec_connections_list(xnm);
	GList* cur;
	int i = 0;
	for(cur = conn_list; cur != NULL; cur = cur->next){
	    gchar *data = g_string_free(((GString*)cur->data), FALSE);
    	gtk_combo_box_append_text (GTK_COMBO_BOX (widget), data);
    	if(g_strcmp0 (value, data) == 0){
        	gtk_combo_box_set_active (GTK_COMBO_BOX (widget), i);
    	}
    	i++;
    }
	if (gtk_combo_box_get_active (GTK_COMBO_BOX (widget)) == -1)
	{
		gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 0);
	}
}

static void
xnm_toggled_cb (GtkCheckButton *button, OpenswanPluginUiWidget *self)
{
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	GtkWidget *widget;
	widget = glade_xml_get_widget (priv->xml, "connname-cb");
	g_assert (widget);
	// reload
	load_connections_list (self, (char *) gtk_combo_box_get_active_text (GTK_COMBO_BOX (widget)));
	g_signal_emit_by_name (self, "changed");	
}

static void
fill_password (GtkWidget *widget,
               NMConnection *connection,
               const char *password_type)
{
	gchar *password = NULL;
	g_assert (widget);

	if (!connection)
		return;

	password = NULL;

	if (nm_connection_get_scope (connection) == NM_CONNECTION_SCOPE_SYSTEM) {
		NMSettingVPN *s_vpn;

		s_vpn = (NMSettingVPN *) nm_connection_get_setting (connection, NM_TYPE_SETTING_VPN);
		if (s_vpn) {
			const gchar *tmp = NULL;

			tmp = nm_setting_vpn_get_secret (s_vpn, password_type);
			if (tmp)
				password = gnome_keyring_memory_strdup (tmp);
		}
	} else {
		NMSettingConnection *s_con = NULL;
		gboolean unused;
		const char *uuid;

		s_con = NM_SETTING_CONNECTION (nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION));
		uuid = nm_setting_connection_get_uuid (s_con);
		password = keyring_helpers_lookup_secret (uuid,
		                                          password_type,
		                                          &unused);
	}

	if (password) {
		gtk_entry_set_text (GTK_ENTRY (widget), password);
		gnome_keyring_memory_free (password);
	}
}

static gboolean
init_plugin_ui (OpenswanPluginUiWidget *self, NMConnection *connection, GError **error)
{
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	NMSettingVPN *settings;
	GtkWidget *widget;
	const char *value;
	
	settings = NM_SETTING_VPN(nm_connection_get_setting(connection, NM_TYPE_SETTING_VPN));
	
	// type-cb init
	widget = glade_xml_get_widget (priv->xml, "type-cb");
	value = nm_setting_vpn_get_data_item (settings, OPENSWAN_PLUGIN_KEY_TYPE);
	
	if(!g_strcmp0(value, OPENSWAN_PLUGIN_TYPE_HOSTTOHOST)){
	    gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 0);
	} else if(!g_strcmp0(value, OPENSWAN_PLUGIN_TYPE_HOSTTONETWORK)){
	    gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 1);
	} else if(!g_strcmp0(value, OPENSWAN_PLUGIN_TYPE_MANUALIPSEC)) {
	    gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 2);
	} else {
        gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 0);
	}
		
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (type_changed_cb), self);
	update_layout (widget, priv);
	
	// xnm-chb-init
	widget = glade_xml_get_widget (priv->xml, "xnm-chb");
	value = nm_setting_vpn_get_data_item (settings, OPENSWAN_PLUGIN_KEY_XNM);
	gboolean xnm = (value && !strcmp (value, "yes"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), xnm);
		
	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (xnm_toggled_cb), self);
	
	// connname-cb init
	widget = glade_xml_get_widget (priv->xml, "connname-cb");
	value = nm_setting_vpn_get_data_item (settings, OPENSWAN_PLUGIN_KEY_CONNECTION_NAME);
	
	load_connections_list(self, value);	
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (settings_changed_cb), self);
	
	
	// gateway-entry init
	widget = glade_xml_get_widget (priv->xml, "gateway-entry");
	value = nm_setting_vpn_get_data_item (settings, OPENSWAN_PLUGIN_KEY_GATEWAY);
	if(value != NULL) {
	    gtk_entry_set_text(GTK_ENTRY(widget), value);
    }
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (settings_changed_cb), self);
		
	// remotenetwork-entry init
	widget = glade_xml_get_widget (priv->xml, "remotenetwork-entry");
	value = nm_setting_vpn_get_data_item (settings, OPENSWAN_PLUGIN_KEY_REMOTE_NETWORK);
	if(value != NULL){
    	gtk_entry_set_text(GTK_ENTRY(widget), value);
    }
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (settings_changed_cb), self);
		
	// mask-entry init
	widget = glade_xml_get_widget (priv->xml, "mask-entry");
	value = nm_setting_vpn_get_data_item (settings, OPENSWAN_PLUGIN_KEY_REMOTE_MASK);
	if(value !=  NULL){
    	gtk_entry_set_text(GTK_ENTRY(widget), value);
    }
	g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (settings_changed_cb), self);
		
	// auth-cb init
	widget = glade_xml_get_widget (priv->xml, "auth-cb");
	value = nm_setting_vpn_get_data_item (settings, OPENSWAN_PLUGIN_KEY_AUTH_TYPE);
	
	if(priv->supports_nss){
	    gtk_combo_box_append_text(GTK_COMBO_BOX (widget), "X.509 (NSS)");
	    gtk_combo_box_append_text(GTK_COMBO_BOX (widget), "Raw RSA (NSS)");	    

	    if(!g_strcmp0(value, OPENSWAN_PLUGIN_AUTH_X509)){
	        gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 0);
	    } else if(!g_strcmp0(value, OPENSWAN_PLUGIN_AUTH_RAWRSA)){
	        gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 1);
	    } else {
            gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 0);
	    }
	    
	    gtk_widget_destroy(glade_xml_get_widget (priv->xml, "cert-label"));
        gtk_widget_destroy(glade_xml_get_widget (priv->xml, "cert-filechooser"));
        gtk_widget_destroy(glade_xml_get_widget (priv->xml, "rsa-label"));
        gtk_widget_destroy(glade_xml_get_widget (priv->xml, "rsa-filechooser"));
        gtk_widget_destroy(glade_xml_get_widget (priv->xml, "password-label"));
        gtk_widget_destroy(glade_xml_get_widget (priv->xml, "password-entry")); 
        gtk_widget_destroy(glade_xml_get_widget (priv->xml, "showpass-chb")); 
        
        g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (settings_changed_cb), self);
	} else {
    	gtk_combo_box_append_text(GTK_COMBO_BOX (widget), "X.509");
	    gtk_combo_box_append_text(GTK_COMBO_BOX (widget), "Raw RSA");
	    gtk_combo_box_append_text(GTK_COMBO_BOX (widget), "PSK");

	    if(!g_strcmp0(value, OPENSWAN_PLUGIN_AUTH_X509)){
	        gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 0);
	    } else if(!g_strcmp0(value, OPENSWAN_PLUGIN_AUTH_RAWRSA)){
	        gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 1);
	    } else if(!g_strcmp0(value, OPENSWAN_PLUGIN_AUTH_PSK)){
	        gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 2);
	    } else {
            gtk_combo_box_set_active (GTK_COMBO_BOX (widget), 0);
	    }
	    g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (auth_changed_cb), self);
        update_layout (widget, priv);
        
        // cert-filechooser init
        widget = glade_xml_get_widget (priv->xml, "cert-filechooser");
        value = nm_setting_vpn_get_data_item (settings, OPENSWAN_PLUGIN_KEY_CERT_FILENAME);
        if(value != NULL){
            gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(widget), value);
        }
        g_signal_connect (G_OBJECT (widget), "selection-changed", G_CALLBACK (settings_changed_cb), self);
        
        // rsa-filechooser init
        widget = glade_xml_get_widget (priv->xml, "rsa-filechooser");
        value = nm_setting_vpn_get_data_item (settings, OPENSWAN_PLUGIN_KEY_RSA_FILENAME);
        if(value != NULL){
            gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(widget), value);
        }
        g_signal_connect (G_OBJECT (widget), "selection-changed", G_CALLBACK (settings_changed_cb), self);
        
        // password-entry init
	    widget = glade_xml_get_widget (priv->xml, "password-entry");
        fill_password (widget, connection, OPENSWAN_PLUGIN_KEY_PASSWORD);
        //value = nm_setting_vpn_get_secret (settings, OPENSWAN_PLUGIN_KEY_PASSWORD);
        //if(value != NULL){
	    //    gtk_entry_set_text(GTK_ENTRY(widget), value);
        //}
	    g_signal_connect (G_OBJECT (widget), "changed", G_CALLBACK (settings_changed_cb), self);
	
	    // showpass-chb init
	    widget = glade_xml_get_widget (priv->xml, "showpass-chb");
	    g_return_val_if_fail (widget != NULL, FALSE);
	    g_signal_connect (G_OBJECT (widget), "toggled",
	                      (GCallback) show_toggled_cb,
	                      self);
	}
		                  
    // advanced-button	
	widget = glade_xml_get_widget (priv->xml, "advanced-button");
    g_signal_connect (G_OBJECT (widget), "clicked", G_CALLBACK (advanced_button_clicked_cb), self);
		               
	return TRUE;
}

static GObject *
get_widget (NMVpnPluginUiWidgetInterface *iface)
{
	OpenswanPluginUiWidget *self = OPENSWAN_PLUGIN_UI_WIDGET (iface);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);

	return G_OBJECT (priv->widget);
}

static void
hash_copy_advanced (gpointer key, gpointer data, gpointer user_data)
{
	NMSettingVPN *s_vpn = NM_SETTING_VPN (user_data);

	nm_setting_vpn_add_data_item (s_vpn, (const char *) key, (const char *) data);
}

static gboolean
update_connection (NMVpnPluginUiWidgetInterface *iface,
                   NMConnection *connection,
                   GError **error)
{
	OpenswanPluginUiWidget *self = OPENSWAN_PLUGIN_UI_WIDGET (iface);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	NMSettingVPN *settings;
	GtkWidget *widget;
	gint active;
	const char *str;

	if (!check_validity (self, error))
		return FALSE;
	settings = NM_SETTING_VPN (nm_setting_vpn_new ());
	
	g_object_set (settings, NM_SETTING_VPN_SERVICE_TYPE,
				  NM_DBUS_SERVICE_OPENSWAN, NULL);
	
	//FIXME: bad idea to save info about nss-support in sysconf
    nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_NSS, priv->supports_nss ? "yes" : "no");

	// type-cb save
	widget = glade_xml_get_widget (priv->xml, "type-cb");
	active = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
	switch (active) {
	case 0: nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_TYPE, OPENSWAN_PLUGIN_TYPE_HOSTTOHOST);
	        break;
	case 1: nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_TYPE, OPENSWAN_PLUGIN_TYPE_HOSTTONETWORK);
	        break;
	case 2: nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_TYPE, OPENSWAN_PLUGIN_TYPE_MANUALIPSEC);
	        break;
	default: nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_TYPE, OPENSWAN_PLUGIN_TYPE_HOSTTOHOST);
	}
	
	// xnm-chb save
	widget = glade_xml_get_widget (priv->xml, "xnm-chb");
	str = (char *) gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)) ? g_strdup ("yes") : g_strdup ("no");
	if (str && strlen (str)) {
		nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_XNM, str);
	}
	
	// connname-cb save
	widget = glade_xml_get_widget (priv->xml, "connname-cb");
	str = (char *) gtk_combo_box_get_active_text (GTK_COMBO_BOX (widget));
	if (str && strlen (str)) {
		nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_CONNECTION_NAME, str);
	}
	
	// gateway-entry save
	widget = glade_xml_get_widget (priv->xml, "gateway-entry");
	str = (char *)gtk_entry_get_text(GTK_ENTRY(widget));
	nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_GATEWAY, str);
	
	// remotenetwork-entry save
	widget = glade_xml_get_widget (priv->xml, "remotenetwork-entry");
	str = (char *)gtk_entry_get_text(GTK_ENTRY(widget));
	nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_REMOTE_NETWORK, str);	
	
	// mask-entry save	                   	         
	widget = glade_xml_get_widget (priv->xml, "mask-entry");     
	str = (char *)gtk_entry_get_text(GTK_ENTRY(widget));
	nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_REMOTE_MASK, str);
	
	// auth-cb save
	widget = glade_xml_get_widget (priv->xml, "auth-cb");
	active = gtk_combo_box_get_active (GTK_COMBO_BOX (widget));
	switch (active) {
	case 1: nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_AUTH_TYPE, OPENSWAN_PLUGIN_AUTH_RAWRSA);
	        break;
	case 2: nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_AUTH_TYPE, OPENSWAN_PLUGIN_AUTH_PSK);
	        break;
	default: nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_AUTH_TYPE, OPENSWAN_PLUGIN_AUTH_X509);
	}
	
	if(!priv->supports_nss){
        // cert-filechooser save
        widget = glade_xml_get_widget (priv->xml, "cert-filechooser");
        str = (char *)gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(widget));
        nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_CERT_FILENAME, str);
        
        // rsa-filechooser save
        widget = glade_xml_get_widget (priv->xml, "rsa-filechooser");
        str = (char *)gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(widget));
        nm_setting_vpn_add_data_item (settings, OPENSWAN_PLUGIN_KEY_RSA_FILENAME, str);
        
        // password-entry save
        //widget = glade_xml_get_widget (priv->xml, "password-entry");
        //str = (char *)gtk_entry_get_text (GTK_ENTRY (widget));
        //nm_setting_vpn_add_secret (settings, OPENSWAN_PLUGIN_KEY_PASSWORD, str);
    }
    
	if (priv->advanced)
		g_hash_table_foreach (priv->advanced, hash_copy_advanced, settings);
	
	nm_connection_add_setting (connection, NM_SETTING (settings));

	return TRUE;
}

static gboolean
save_secrets (NMVpnPluginUiWidgetInterface *iface,
              NMConnection *connection, GError **error)
{
	OpenswanPluginUiWidget *self = OPENSWAN_PLUGIN_UI_WIDGET (iface);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (self);
	if(priv->supports_nss)
	    return TRUE;
	GnomeKeyringResult ret;
	NMSettingConnection *s_con;
	GtkWidget *widget;
	const char *str, *uuid, *id;
    
    NMSettingVPN* settings;
    settings = NM_SETTING_VPN(nm_connection_get_setting(connection, NM_TYPE_SETTING_VPN));
    
    //widget = glade_xml_get_widget (priv->xml, "password-entry");
    //if(widget){
    //    str = (char *)gtk_entry_get_text (GTK_ENTRY (widget));
    //    nm_setting_vpn_add_secret (settings, OPENSWAN_PLUGIN_KEY_PASSWORD, str);
    //}
	
	s_con = (NMSettingConnection *) nm_connection_get_setting (connection, NM_TYPE_SETTING_CONNECTION);
	if (!s_con) {
		g_set_error (error,
		             OPENSWAN_PLUGIN_UI_ERROR,
		             OPENSWAN_PLUGIN_UI_ERROR_INVALID_CONNECTION,
		             "missing 'connection' setting");
		return FALSE;
	}

	id = nm_setting_connection_get_id (s_con);
	uuid = nm_setting_connection_get_uuid (s_con);

    widget = glade_xml_get_widget (priv->xml, "password-entry");
    g_assert (widget);
    str = gtk_entry_get_text (GTK_ENTRY (widget));
    if (str && strlen (str)) {
        ret = keyring_helpers_save_secret (uuid, id, NULL, OPENSWAN_PLUGIN_KEY_PASSWORD, str);
        if (ret != GNOME_KEYRING_RESULT_OK)
            g_warning ("%s: failed to save user password to keyring.", __func__);
    } else
        keyring_helpers_delete_secret (uuid, OPENSWAN_PLUGIN_KEY_PASSWORD);
    
	return TRUE;
}

static gboolean
delete_connection (NMVpnPluginUiInterface *iface,
                   NMConnection *connection,
                   GError **error)
{
	NMSettingConnection *s_con = NULL;
	const char *uuid;

	/* Remove any secrets in the keyring associated with this connection's UUID */
	s_con = (NMSettingConnection *) nm_connection_get_setting (connection,
			NM_TYPE_SETTING_CONNECTION);
	if (!s_con) {
		g_set_error (error,
		             OPENSWAN_PLUGIN_UI_ERROR,
		             OPENSWAN_PLUGIN_UI_ERROR_INVALID_CONNECTION,
		             "missing 'connection' setting");
		return FALSE;
	}

	uuid = nm_setting_connection_get_uuid (s_con);
	keyring_helpers_delete_secret (uuid, OPENSWAN_PLUGIN_KEY_PASSWORD);

	return TRUE;
}

static NMVpnPluginUiWidgetInterface *
nm_vpn_plugin_ui_widget_interface_new (NMConnection *connection, GError **error)
{
	NMVpnPluginUiWidgetInterface *object;
	OpenswanPluginUiWidgetPrivate *priv;
	char *glade_file;

	if (error)
		g_return_val_if_fail (*error == NULL, NULL);
    
	object = NM_VPN_PLUGIN_UI_WIDGET_INTERFACE (g_object_new (OPENSWAN_TYPE_PLUGIN_UI_WIDGET, NULL));
	if (!object) {
		g_set_error (error, OPENSWAN_PLUGIN_UI_ERROR, 0, "could not create openswan object");
		return NULL;
	}

	priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (object);

    priv->supports_nss = is_openswan_support_nss();
    
	glade_file = g_strdup_printf ("%s/%s", GLADEDIR, "nm-openswan-dialog.glade");
	priv->xml = glade_xml_new (glade_file, "openswan-vbox", GETTEXT_PACKAGE);
	if (priv->xml == NULL) {
		g_set_error (error, OPENSWAN_PLUGIN_UI_ERROR, 0,
		             "could not load required resources at %s", glade_file);
		g_free (glade_file);
		g_object_unref (object);
		return NULL;
	}
	g_free (glade_file);

	priv->widget = glade_xml_get_widget (priv->xml, "openswan-vbox");
	if (!priv->widget) {
		g_set_error (error, OPENSWAN_PLUGIN_UI_ERROR, 0, "could not load UI widget");
		g_object_unref (object);
		return NULL;
	}
	g_object_ref_sink (priv->widget);

	priv->window_group = gtk_window_group_new ();
	
	priv->left_size_group = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
	priv->right_size_group = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);

	gtk_size_group_add_widget(priv->right_size_group, glade_xml_get_widget (priv->xml, "type-cb"));
	gtk_size_group_add_widget(priv->right_size_group, glade_xml_get_widget (priv->xml, "gateway-entry"));
	gtk_size_group_add_widget(priv->right_size_group, glade_xml_get_widget (priv->xml, "remotenetwork-entry"));
	gtk_size_group_add_widget(priv->right_size_group, glade_xml_get_widget (priv->xml, "mask-entry"));
	gtk_size_group_add_widget(priv->right_size_group, glade_xml_get_widget (priv->xml, "connname-cb"));
	gtk_size_group_add_widget(priv->right_size_group, glade_xml_get_widget (priv->xml, "auth-cb"));
	gtk_size_group_add_widget(priv->right_size_group, glade_xml_get_widget (priv->xml, "cert-filechooser"));
	gtk_size_group_add_widget(priv->right_size_group, glade_xml_get_widget (priv->xml, "rsa-filechooser"));
	gtk_size_group_add_widget(priv->right_size_group, glade_xml_get_widget (priv->xml, "password-entry"));
	
	gtk_size_group_add_widget(priv->left_size_group, glade_xml_get_widget (priv->xml, "type-label"));
	gtk_size_group_add_widget(priv->left_size_group, glade_xml_get_widget (priv->xml, "gateway-label"));
	gtk_size_group_add_widget(priv->left_size_group, glade_xml_get_widget (priv->xml, "remotenetwork-label"));
	gtk_size_group_add_widget(priv->left_size_group, glade_xml_get_widget (priv->xml, "mask-label"));
	gtk_size_group_add_widget(priv->left_size_group, glade_xml_get_widget (priv->xml, "connname-label"));
	gtk_size_group_add_widget(priv->left_size_group, glade_xml_get_widget (priv->xml, "auth-label"));
	gtk_size_group_add_widget(priv->left_size_group, glade_xml_get_widget (priv->xml, "cert-label"));
	gtk_size_group_add_widget(priv->left_size_group, glade_xml_get_widget (priv->xml, "rsa-label"));
	gtk_size_group_add_widget(priv->left_size_group, glade_xml_get_widget (priv->xml, "password-label"));

	if (!init_plugin_ui (OPENSWAN_PLUGIN_UI_WIDGET (object), connection, error)) {
		g_object_unref (object);
		return NULL;
	}

	priv->advanced = advanced_dialog_new_hash_from_connection (connection, error);
	if (!priv->advanced) {
		g_object_unref (object);
		return NULL;
	}
    
    g_printf("nss? %i\n", priv->supports_nss);

	return object;
}

static void
dispose (GObject *object)
{
	OpenswanPluginUiWidget *plugin = OPENSWAN_PLUGIN_UI_WIDGET (object);
	OpenswanPluginUiWidgetPrivate *priv = OPENSWAN_PLUGIN_UI_WIDGET_GET_PRIVATE (plugin);

	if (priv->widget)
		g_object_unref (priv->widget);

	if (priv->xml)
		g_object_unref (priv->xml);
		
    if (priv->window_group)
		g_object_unref (priv->window_group);
	
	if (priv->advanced)
		g_hash_table_destroy (priv->advanced);

	G_OBJECT_CLASS (openswan_plugin_ui_widget_parent_class)->dispose (object);
}

static void
openswan_plugin_ui_widget_class_init (OpenswanPluginUiWidgetClass *req_class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (req_class);

	g_type_class_add_private (req_class, sizeof (OpenswanPluginUiWidgetPrivate));

	object_class->dispose = dispose;
}

static void
openswan_plugin_ui_widget_init (OpenswanPluginUiWidget *plugin)
{
}

static void
openswan_plugin_ui_widget_interface_init (NMVpnPluginUiWidgetInterface *iface_class)
{
	/* interface implementation */
	iface_class->get_widget = get_widget;
	iface_class->update_connection = update_connection;
	//if(!ENABLE_NSS)
    iface_class->save_secrets = save_secrets;
}

static guint32
get_capabilities (NMVpnPluginUiInterface *iface)
{
	return 0;
}

static NMVpnPluginUiWidgetInterface *
ui_factory (NMVpnPluginUiInterface *iface, NMConnection *connection, GError **error)
{
	return nm_vpn_plugin_ui_widget_interface_new (connection, error);
}

static void
get_property (GObject *object, guint prop_id,
			  GValue *value, GParamSpec *pspec)
{
	switch (prop_id) {
	case NM_VPN_PLUGIN_UI_INTERFACE_PROP_NAME:
		g_value_set_string (value, OPENSWAN_PLUGIN_NAME);
		break;
	case NM_VPN_PLUGIN_UI_INTERFACE_PROP_DESC:
		g_value_set_string (value, OPENSWAN_PLUGIN_DESC);
		break;
	case NM_VPN_PLUGIN_UI_INTERFACE_PROP_SERVICE:
		g_value_set_string (value, OPENSWAN_PLUGIN_SERVICE);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
openswan_plugin_ui_class_init (OpenswanPluginUiClass *req_class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (req_class);

	object_class->get_property = get_property;

	g_object_class_override_property (object_class,
									  NM_VPN_PLUGIN_UI_INTERFACE_PROP_NAME,
									  NM_VPN_PLUGIN_UI_INTERFACE_NAME);

	g_object_class_override_property (object_class,
									  NM_VPN_PLUGIN_UI_INTERFACE_PROP_DESC,
									  NM_VPN_PLUGIN_UI_INTERFACE_DESC);

	g_object_class_override_property (object_class,
									  NM_VPN_PLUGIN_UI_INTERFACE_PROP_SERVICE,
									  NM_VPN_PLUGIN_UI_INTERFACE_SERVICE);
}

static void
openswan_plugin_ui_init (OpenswanPluginUi *plugin)
{
}

static void
openswan_plugin_ui_interface_init (NMVpnPluginUiInterface *iface_class)
{
	/* interface implementation */
	iface_class->ui_factory = ui_factory;
	iface_class->get_capabilities = get_capabilities;
	iface_class->delete_connection = delete_connection;

}


G_MODULE_EXPORT NMVpnPluginUiInterface *
nm_vpn_plugin_ui_factory (GError **error)
{
	if (error)
		g_return_val_if_fail (*error == NULL, NULL);

	return NM_VPN_PLUGIN_UI_INTERFACE (g_object_new (OPENSWAN_TYPE_PLUGIN_UI, NULL));
}

